#!/usr/bin/env python3

import pytest
import autoprop

def test_keyword_only_arguments():
    # inspect.getargspec() chokes methods with keyword-only arguments.

    @autoprop   # (no fold)
    class Example(object):
        def get_attr(self, *, pos=None):
            return 'attr'
        def set_attr(self, *, new_value):
            pass

    ex = Example()
    assert ex.attr == 'attr'

    # Even though the setter has one required argument, it's a named argument, 
    # so it shouldn't count and 'attr' should be defined without a setter.
    with pytest.raises(AttributeError):
        ex.attr = 'setter'





#!/usr/bin/env python3

import sys

def pytest_ignore_collect(path, config):
    if sys.version_info[0] == 2:
        return 'python3' in str(path)
    if sys.version_info[0] == 3:
        return 'python2' in str(path)



#!/usr/bin/env python2

import pytest
import autoprop

def test_complain_about_old_style_classes():
    with pytest.raises(TypeError) as err:
        @autoprop
        class Example:
            pass

    assert '@autoprop can only be used with new-style classes' in str(err)




#!/usr/bin/env python

import pytest
import autoprop

def test_get():
    @autoprop
    class Example(object): #
        def get_attr(self): #
            return 'attr'

    ex = Example()
    assert ex.attr == 'attr'

def test_set():
    @autoprop
    class Example(object): #
        def set_attr(self, attr): #
            self._attr = 'new ' + attr

    ex = Example()
    ex.attr = 'attr'
    assert ex._attr == 'new attr'

def test_del():
    @autoprop
    class Example(object): #
        def del_attr(self): #
            self._attr = None

    ex = Example()
    del ex.attr
    assert ex._attr is None

def test_get_set():
    @autoprop
    class Example(object): #
        def get_attr(self): #
            return self._attr
        def set_attr(self, attr): #
            self._attr = 'new ' + attr

    ex = Example()
    ex.attr = 'attr'
    assert ex.attr == 'new attr'

def test_get_set_del():
    @autoprop
    class Example(object): #
        def get_attr(self): #
            return self._attr
        def set_attr(self, attr): #
            self._attr = 'new ' + attr
        def del_attr(self): #
            self._attr = None

    ex = Example()
    ex.attr = 'attr'
    assert ex.attr == 'new attr'
    del ex.attr
    assert ex.attr is None

def test_cache_get():

    @autoprop
    class Example(object):

        def __init__(self, attr): #
            self._attr = attr

        @autoprop.cache
        def get_attr(self): #
            self._attr += 1
            return self._attr

    ex1 = Example(0)
    ex2 = Example(2)

    # Calling the getter multiple times only call the underlying method once.
    assert ex1.attr == 1
    assert ex1.attr == 1

    # Different objects are independent of each other.
    assert ex2.attr == 3
    assert ex1.attr == 1
    assert ex2.attr == 3
    assert ex1.attr == 1

def test_cache_get_set_del():

    @autoprop
    class Example(object):

        def __init__(self, attr): #
            self._attr = attr

        @autoprop.cache
        def get_attr(self): #
            self._attr += 1
            return self._attr

        def set_attr(self, attr): #
            self._attr = attr

        def del_attr(self): #
            self._attr = 0

    ex1 = Example(0)
    ex2 = Example(2)

    # Calling the getter multiple times only call the underlying method once:
    assert ex1.attr == 1
    assert ex1.attr == 1

    assert ex2.attr == 3
    assert ex2.attr == 3

    # Calling the setter clears the cache without affecting other instances:
    ex1.attr = 4
    assert ex1.attr == 5
    assert ex1.attr == 5

    assert ex2.attr == 3
    assert ex2.attr == 3

    # Calling the deleter clears the cache without affecting other instances:
    del ex1.attr
    assert ex1.attr == 1
    assert ex1.attr == 1

    assert ex2.attr == 3
    assert ex2.attr == 3

def test_dont_cache_non_getter():
    with pytest.raises(ValueError, match=r"not_a_getter\(\) cannot be cached"):

        @autoprop
        class Example:

            @autoprop.cache
            def not_a_getter(self):
                pass

def test_cache_inheritance():
    # The child class determines whether or not the attribute is cached.

    @autoprop
    class Parent(object):

        def __init__(self, v=0): #
            self._w = v
            self._x = v
            self._y = v
            self._z = v

        @autoprop.cache
        def get_super_cache(self):
            self._w += 1
            return self._w

        @autoprop.cache
        def get_parent_cache(self): #
            self._x += 1
            return self._x

        def get_child_cache(self): #
            self._y += 1
            return self._y

        @autoprop.cache
        def get_both_cache(self): #
            self._z += 1
            return self._z

    @autoprop
    class Child(Parent): #

        def get_parent_cache(self): #
            self._x += 1
            return self._x

        @autoprop.cache
        def get_child_cache(self): #
            self._y += 1
            return self._y

        @autoprop.cache
        def get_both_cache(self): #
            self._z += 1
            return self._z

    p = Parent()
    c = Child(10)

    assert p.parent_cache == 1
    assert p.parent_cache == 1
    assert c.parent_cache == 11
    assert c.parent_cache == 12

    assert p.child_cache == 1
    assert p.child_cache == 2
    assert c.child_cache == 11
    assert c.child_cache == 11

    assert p.both_cache == 1
    assert p.both_cache == 1
    assert c.both_cache == 11
    assert c.both_cache == 11

    assert p.super_cache == 1
    assert p.super_cache == 1
    assert c.super_cache == 11
    assert c.super_cache == 11

def test_ignore_similar_names():
    @autoprop #
    class Example(object):
        def getattr(self):
            return 'attr'

    ex = Example()
    with pytest.raises(AttributeError):
        ex.attr

def test_ignore_empty_names():
    @autoprop #
    class Example(object):
        def get_(self):
            return 'get'

    ex = Example()
    with pytest.raises(AttributeError):
        getattr(ex, '')

def test_ignore_non_methods():
    @autoprop #
    class Example(object):
        get_attr = 'attr'

    ex = Example()
    with pytest.raises(AttributeError):
        ex.attr

def test_dont_overwrite_existing_attributes():
    @autoprop #
    class Example(object):
        attr = 'class var'
        def get_attr(self):
            return 'attr'

    ex = Example()
    assert ex.attr == 'class var'

def test_dont_overwrite_inherited_attributes():
    @autoprop #
    class Parent(object):
        attr = 'class var'
        def get_attr(self):
            return 'parent'

    @autoprop #
    class Child(Parent):
        def get_attr(self):
            return 'child'

    parent = Parent()
    child = Child()

    assert parent.attr == 'class var'
    assert child.attr == 'class var'

def test_overwrite_inherited_autoprops():
    @autoprop
    class Parent(object): #
        def get_attr(self): #
            return 'parent'
        def get_overloaded_attr(self): #
            return 'parent'

    @autoprop
    class Child(Parent): #
        def get_overloaded_attr(self): #
            return 'child'

    parent = Parent()
    child = Child()

    assert parent.attr == 'parent'
    assert parent.overloaded_attr == 'parent'
    assert child.attr == 'parent'
    assert child.overloaded_attr == 'child'

def test_optional_arguments():
    @autoprop #
    class Example(object):
        def get_attr(self, pos=None, *args, **kwargs):
            return self._attr
        def set_attr(self, new_value, pos=None, *args, **kwargs):
            self._attr = 'new ' + new_value
        def del_attr(self, pos=None, *args, **kwargs):
            self._attr = None

    ex = Example()
    ex.attr = 'attr'
    assert ex.attr == 'new attr'
    del ex.attr
    assert ex.attr == None

def test_getters_need_one_argument():
    @autoprop #
    class Example(object):
        def get_attr():
            return 'attr'

    ex = Example()
    with pytest.raises(AttributeError):
        ex.attr

    @autoprop #
    class Example(object):
        def get_attr(self, more_info):
            return 'attr'

    ex = Example()
    with pytest.raises(AttributeError):
        ex.attr

def test_setters_need_two_arguments():
    @autoprop #
    class Example(object):
        def set_attr(self):
            self._attr = 'no args'

    ex = Example()
    ex.attr = 'attr'
    with pytest.raises(AttributeError):
        assert ex._attr != 'no args'

    @autoprop #
    class Example(object):
        def set_attr(self, new_value, more_info):
            self._attr = 'two args'

    ex = Example()
    ex.attr = 'attr'
    with pytest.raises(AttributeError):
        assert ex._attr != 'two args'

def test_deleters_need_one_argument():
    @autoprop #
    class Example(object):
        def del_attr():
            pass

    ex = Example()
    with pytest.raises(AttributeError):
        del ex.attr

    @autoprop #
    class Example(object):
        def del_attr(self, more_info):
            pass

    ex = Example()
    with pytest.raises(AttributeError):
        del ex.attr

def test_docstrings():

    @autoprop #
    class Example(object):
        def get_attr(self): #
            "get attr"
            return self._attr
        def set_attr(self, attr): #
            "set attr"
            self._attr = 'new ' + attr
        def del_attr(self): #
            "del attr"
            self._attr = None

    # The docstrings on the setter and deleter are ignored, as per the default 
    # behavior of the `property()` decorator.
    assert Example.attr.__doc__ == "get attr"



#!/usr/bin/env python

class autoprop(object):

    __version__ = '1.0.1'
    property = type('property', (property,), {})

    class _Cache:
        __slots__ = ('value', 'is_stale')

        def __init__(self):
            self.value = None
            self.is_stale = True

        def __repr__(self):
            import sys
            return '{}(value={!r}, is_stale={!r})'.format(
                    self.__class__.__qualname__ if sys.version_info[0] >= 3 else self.__class__.__name__,
                    self.value,
                    self.is_stale,
            )

    def __new__(autoprop, cls):
        # These imports have to be inside autoprop(), otherwise the sys.modules 
        # hack below somehow makes them unavailable when the decorator is 
        # applied.
        import inspect, re
        from collections import defaultdict
        from functools import wraps

        if not hasattr(cls, '__class__'):
            raise TypeError('@autoprop can only be used with new-style classes')

        accessors = defaultdict(dict)
        expected_num_args = {'get': 0, 'set': 1, 'del': 0}

        # The accessors we're searching for are considered methods in python2 
        # and functions in python3.  They behave the same either way.
        ismethod = lambda x: inspect.ismethod(x) or inspect.isfunction(x)

        for method_name, method in inspect.getmembers(cls, ismethod):
            accessor_match = re.match('(get|set|del)_(.+)', method_name)
            if not accessor_match:
                continue

            # Suppress a warning by using getfullargspec() if it's available 
            # and getargspec() if it's not.
            try: from inspect import getfullargspec as getargspec
            except ImportError: from inspect import getargspec

            prefix, name = accessor_match.groups()
            arg_spec = getargspec(method)
            num_args = len(arg_spec.args) - len(arg_spec.defaults or ())
            num_args_minus_self = num_args - 1

            if num_args_minus_self != expected_num_args[prefix]:
                continue

            accessors[name][prefix] = method

        def get_cache(self, getter):
            if not hasattr(self, '_autoprop_cache'):
                self._autoprop_cache = {}
            return self._autoprop_cache.setdefault(getter, autoprop._Cache())

        def use_cache(getter):

            @wraps(getter)
            def wrapper(self, *args, **kwargs):
                cache = get_cache(self, getter)

                if cache.is_stale:
                    cache.value = getter(self, *args, **kwargs)
                    cache.is_stale = False

                return cache.value

            return wrapper

        def clear_cache(f, getter):
            if f is None:
                return None

            @wraps(f)
            def wrapper(self, *args, **kwargs):
                get_cache(self, getter).is_stale = True
                return f(self, *args, **kwargs)

            return wrapper

        for name in accessors:
            try:
                attr = getattr(cls, name)
                ok_to_overwrite = isinstance(attr, autoprop.property)
            except AttributeError:
                ok_to_overwrite = True

            getter  = accessors[name].get('get')
            setter  = accessors[name].get('set')
            deleter = accessors[name].get('del')

            # Cache the return value of the getter, if requested.  For some 
            # reason setting attributes on method objects 
            if getter and getattr(getter, '_autoprop_mark_cache', False):
                setter  = clear_cache(setter,  getter)
                deleter = clear_cache(deleter, getter)
                getter  = use_cache(getter)

            if ok_to_overwrite:
                property = autoprop.property(getter, setter, deleter)
                setattr(cls, name, property)

        return cls

    @staticmethod
    def cache(f):
        if not f.__name__.startswith('get_'):
            import sys
            raise ValueError("{}() cannot be cached; it's not a getter".format(
                f.__qualname__ if sys.version_info[0] >= 3 else f.__name__
            ))

        f._autoprop_mark_cache = True
        return f

# Abuse the import system so that the module itself can be used as a decorator.  
# This is a simple module intended only to cut-down on boilerplate, so I think 
# the trade-off between magicalness and ease-of-use is justified in this case.
import sys
sys.modules[__name__] = autoprop



#!/usr/bin/env python3
# encoding: utf-8

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

setup(
    name='autoprop',
    version='1.0.1',
    author='Kale Kundert',
    author_email='kale@thekunderts.net',
    long_description=open('README.rst').read(),
    url='https://github.com/kalekundert/autoprop',
    license='MIT',
    py_modules=[
        'autoprop',
    ],
    keywords=[
        'property', 'properties',
        'accessor', 'accessors',
        'getter', 'setter'
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries',
    ],
)
